@@ -1,20 +1,25 @@
 from __future__ import unicode_literals
 from chameleon.config import AUTO_RELOAD
 from devpi_common.metadata import get_latest_version
-from devpi_web import hookspecs
-from devpi_web.description import render_description
-from devpi_web.doczip import unpack_docs
-from devpi_web.indexing import iter_projects, preprocess_project
-from devpi_web.whoosh_index import Index
+from devpi_web.config import add_indexer_backend_option
+from devpi_web.config import get_pluginmanager
+from devpi_web.doczip import remove_docs
+from devpi_web.indexing import ProjectIndexingInfo
+from devpi_web.indexing import is_project_cached
 from devpi_server.log import threadlog
+from devpi_server.main import fatal
 from pkg_resources import resource_filename
-from pluggy import PluginManager
+from pluggy import HookimplMarker
 from pyramid.renderers import get_renderer
 from pyramid_chameleon.renderer import ChameleonRendererLookup
 import os
 import sys
 
 
+hookimpl = HookimplMarker("devpiweb")
+devpiserver_hookimpl = HookimplMarker("devpiserver")
+
+
 def theme_static_url(request, path):
     return request.static_url(
         os.path.join(request.registry['theme_path'], 'static', path))
@@ -40,10 +45,10 @@ def navigation_version(context):
     version = context.version
     if version == 'latest':
         stage = context.model.getstage(context.username, context.index)
-        version = stage.get_latest_version(context.project)
+        version = stage.get_latest_version_perstage(context.project)
     elif version == 'stable':
         stage = context.model.getstage(context.username, context.index)
-        version = stage.get_latest_version(context.project, stable=True)
+        version = stage.get_latest_version_perstage(context.project, stable=True)
     return version
 
 
@@ -55,13 +60,17 @@ def navigation_info(request):
     result = dict(path=path)
     if context.matchdict and 'user' in context.matchdict:
         user = context.username
+        path.append(dict(
+            url=request.route_url(
+                "/{user}", user=user),
+            title="%s" % user))
     else:
         return result
     if 'index' in context.matchdict:
         index = context.index
         path.append(dict(
             url=request.stage_url(user, index),
-            title="%s/%s" % (user, index)))
+            title="%s" % index))
     else:
         return result
     if 'project' in context.matchdict:
@@ -123,15 +132,6 @@ class ThemeChameleonRendererLookup(ChameleonRendererLookup):
         return ChameleonRendererLookup.__call__(self, info)
 
 
-def get_pluginmanager(load_entry_points=True):
-    pm = PluginManager("devpiweb", implprefix="devpiweb_")
-    pm.add_hookspecs(hookspecs)
-    if load_entry_points:
-        pm.load_setuptools_entrypoints("devpi_web")
-    pm.check_pending()
-    return pm
-
-
 def includeme(config):
     from devpi_web import __version__
     from pyramid_chameleon.interfaces import IChameleonLookup
@@ -170,16 +170,43 @@ def includeme(config):
     config.add_request_method(navigation_info, reify=True)
     config.add_request_method(status_info, reify=True)
     config.add_request_method(query_docs_html, reify=True)
-    config.registry['devpiweb-pluginmanager'] = get_pluginmanager()
     config.scan()
 
 
-def get_indexer(config):
-    indices_dir = config.serverdir.join('.indices')
-    indices_dir.ensure_dir()
-    return Index(indices_dir.strpath)
+def get_indexer_from_config(config):
+    pm = get_pluginmanager(config)
+    indexers = {
+        x['name']: x
+        for x in pm.hook.devpiweb_indexer_backend()}
+    if isinstance(config.args.indexer_backend, dict):
+        # a yaml config may return a dict
+        settings = dict(config.args.indexer_backend)
+        name = settings.pop('name')
+    else:
+        (name, sep, setting_str) = config.args.indexer_backend.partition(':')
+        settings = {}
+        if setting_str:
+            for item in setting_str.split(','):
+                (key, value) = item.split('=', 1)
+                settings[key] = value
+    indexer = indexers[name]['indexer'](config=config, settings=settings)
+    return indexer
 
 
+def get_indexer(xom):
+    # lookup cached value
+    indexer = getattr(xom, 'devpiweb_indexer', None)
+    if indexer is not None:
+        return indexer
+    indexer = get_indexer_from_config(xom.config)
+    if hasattr(indexer, 'runtime_setup'):
+        indexer.runtime_setup(xom)
+    # cache the expensive setup
+    xom.devpiweb_indexer = indexer
+    return indexer
+
+
+@devpiserver_hookimpl
 def devpiserver_pyramid_configure(config, pyramid_config):
     # make the theme path absolute if it exists and make it available via the
     # pyramid registry
@@ -198,7 +225,9 @@ def devpiserver_pyramid_configure(config, pyramid_config):
     # by using include, the package name doesn't need to be set explicitly
     # for registrations of static views etc
     pyramid_config.include('devpi_web.main')
-    pyramid_config.registry['search_index'] = get_indexer(config)
+    pyramid_config.registry['devpiweb-pluginmanager'] = get_pluginmanager(config)
+    xom = pyramid_config.registry['xom']
+    pyramid_config.registry['search_index'] = get_indexer(xom)
 
     # monkeypatch mimetypes.guess_type on because pyramid-1.5.1/webob
     # choke on mimtypes.guess_type on windows with python2.7
@@ -214,47 +243,46 @@ def devpiserver_pyramid_configure(config, pyramid_config):
         threadlog.debug("monkeypatched mimetypes.guess_type to return bytes")
 
 
+@devpiserver_hookimpl
 def devpiserver_add_parser_options(parser):
-    web = None
-    for action in parser._actions:
-        if '--host' in action.option_strings:
-            web = action.container
-    web.addoption(
+    theme = parser.addgroup("devpi-web theme options")
+    theme.addoption(
         "--theme", action="store",
         help="folder with template and resource overwrites for the web interface")
-    indexing = parser.addgroup("search indexing")
-    indexing.addoption(
-        "--recreate-search-index", action="store_true",
-        help="Recreate search index for all projects and their documentation. "
-             "This is only needed if there where indexing related errors in a "
-             "devpi-web release and you want to upgrade only devpi-web "
-             "without a full devpi-server import/export.")
+    doczip = parser.addgroup("devpi-web doczip options")
+    doczip.addoption(
+        "--documentation-path", action="store",
+        help="path for unzipped documentation. "
+             "By default the --serverdir is used.")
+    indexing = parser.addgroup("devpi-web search indexing")
+    add_indexer_backend_option(indexing)
 
 
+@devpiserver_hookimpl
 def devpiserver_mirror_initialnames(stage, projectnames):
-    xom = stage.xom
-    ix = get_indexer(xom.config)
-    ix.delete_index()
-    indexer = get_indexer(xom.config)
-    # directly use projectnames?
-    indexer.update_projects(iter_projects(xom), clear=True)
-    threadlog.info("finished initial indexing op")
+    ix = get_indexer(stage.xom)
+    threadlog.info(
+        "indexing '%s' mirror with %s projects",
+        stage.name,
+        len(projectnames))
+    ix.update_projects(
+        ProjectIndexingInfo(stage=stage, name=name)
+        for name in projectnames)
+    threadlog.info("finished mirror indexing operation")
 
 
+@devpiserver_hookimpl
 def devpiserver_stage_created(stage):
     if stage.ixconfig["type"] == "mirror":
         threadlog.info("triggering load of initial projectnames for %s", stage.name)
         stage.list_projects_perstage()
 
 
+@devpiserver_hookimpl
 def devpiserver_cmdline_run(xom):
-    ix = get_indexer(xom.config)
-    if xom.config.args.recreate_search_index:
-        ix.delete_index()
-        indexer = get_indexer(xom.config)
-        indexer.update_projects(iter_projects(xom), clear=True)
-        # only exit when indexing explicitly
-        return 0
+    docs_path = xom.config.args.documentation_path
+    if docs_path is not None and not os.path.isabs(docs_path):
+        fatal("The path for unzipped documentation must be absolute.")
     # allow devpi-server to run
     return None
 
@@ -262,33 +290,34 @@ def devpiserver_cmdline_run(xom):
 def delete_project(stage, name):
     if stage is None:
         return
-    ix = get_indexer(stage.xom.config)
-    ix.delete_projects([preprocess_project(stage, name)])
+    ix = get_indexer(stage.xom)
+    ix.delete_projects([ProjectIndexingInfo(stage=stage, name=name)])
 
 
 def index_project(stage, name):
     if stage is None:
         return
-    ix = get_indexer(stage.xom.config)
-    ix.update_projects([preprocess_project(stage, name)])
+    ix = get_indexer(stage.xom)
+    ix.update_projects([ProjectIndexingInfo(stage=stage, name=name)])
 
 
+@devpiserver_hookimpl
 def devpiserver_on_upload(stage, project, version, link):
     if not link.entry.file_exists():
         # on replication or import we might be at a lower than
         # current revision and the file might have been deleted already
         threadlog.debug("ignoring lost upload: %s", link)
     elif link.rel == "doczip":
-        unpack_docs(stage, project, version, link.entry)
         index_project(stage, project)
 
 
+@devpiserver_hookimpl
 def devpiserver_on_changed_versiondata(stage, project, version, metadata):
     if stage is None:
         # TODO we don't have enough info to delete the project
         return
     if not metadata:
-        if not stage.has_project_perstage(project):
+        if is_project_cached(stage, project) and not stage.has_project_perstage(project):
             delete_project(stage, project)
             return
         versions = stage.list_versions(project)
@@ -299,5 +328,21 @@ def devpiserver_on_changed_versiondata(stage, project, version, metadata):
                     project, version))
                 metadata = stage.get_versiondata(project, version)
     if metadata:
-        render_description(stage, metadata)
         index_project(stage, metadata['name'])
+
+
+@devpiserver_hookimpl
+def devpiserver_on_remove_file(stage, relpath):
+    if relpath.endswith(".doc.zip"):
+        project, version = (
+            os.path.basename(relpath).rsplit('.doc.zip')[0].rsplit('-', 1)
+        )
+        remove_docs(stage, project, version)
+
+
+@devpiserver_hookimpl(optionalhook=True)
+def devpiserver_on_replicated_file(stage, project, version, link, serial, back_serial, is_from_mirror):
+    if is_from_mirror:
+        return
+    elif link.rel == "doczip":
+        index_project(stage, project)
