@@ -1,6 +1,7 @@
 # coding: utf-8
 from __future__ import unicode_literals
 from defusedxml.xmlrpc import DefusedExpatParser
+from devpi_common.metadata import Version
 from devpi_common.metadata import get_pyversion_filetype
 from devpi_common.metadata import get_sorted_versions
 from devpi_common.validation import normalize_name
@@ -9,7 +10,7 @@ from devpi_server.log import threadlog as log
 from devpi_server.readonly import SeqViewReadonly
 from devpi_server.views import StatusView, url_for_entrypath
 from devpi_web.description import get_description
-from devpi_web.doczip import Docs, get_unpack_path
+from devpi_web.doczip import Docs, unpack_docs
 from devpi_web.indexing import is_project_cached
 from devpi_web.main import navigation_version
 from email.utils import parsedate
@@ -42,50 +43,81 @@ class ContextWrapper(object):
     def __getattr__(self, name):
         return getattr(self.context, name)
 
+    @reify
+    def resolved_version(self):
+        version = self.version
+        if version == 'latest' and self._versions:
+            version = self._versions[0]
+        elif version == 'stable' and self._stable_versions:
+            version = self._stable_versions[0]
+        return version
+
+    @reify
+    def _versions(self):
+        return get_sorted_versions(
+            self.stage.list_versions_perstage(self.project),
+            stable=False)
+
     @reify
     def versions(self):
-        versions = self.stage.list_versions(self.project)
+        versions = self._versions
         if not versions:
             raise HTTPNotFound("The project %s does not exist." % self.project)
-        return get_sorted_versions(versions)
+        return versions
+
+    @reify
+    def _stable_versions(self):
+        return get_sorted_versions(self._versions, stable=True)
 
     @reify
     def stable_versions(self):
-        versions = self.stage.list_versions(self.project)
-        if not versions:
+        if not self._versions:
             raise HTTPNotFound("The project %s does not exist." % self.project)
-        versions = get_sorted_versions(versions, stable=True)
+        versions = self._stable_versions
         if not versions:
             raise HTTPNotFound("The project %s has no stable release." % self.project)
         return versions
 
     @reify
     def linkstore(self):
-        return self.stage.get_linkstore_perstage(self.project, self.version)
+        try:
+            return self.stage.get_linkstore_perstage(self.project, self.version)
+        except self.stage.MissesRegistration:
+            raise HTTPNotFound(
+                "%s-%s is not registered" % (self.project, self.version))
 
 
-def get_doc_info(context, request):
+def get_doc_info(context, request, version=None, check_content=True):
     relpath = request.matchdict['relpath']
     if not relpath:
         raise HTTPFound(location="index.html")
     name = context.project
-    version = context.version
+    if version is None:
+        version = context.version
     if version == 'latest':
         versions = context.versions
     elif version == 'stable':
         versions = context.stable_versions
     else:
         versions = [version]
+    doc_path = None
     for doc_version in versions:
-        doc_path = get_unpack_path(context.stage, name, doc_version)
+        try:
+            linkstore = context.stage.get_linkstore_perstage(name, doc_version)
+        except context.stage.MissesRegistration:
+            continue
+        links = linkstore.get_links(rel='doczip')
+        if not links:
+            continue
+        doc_path = unpack_docs(context.stage, name, doc_version, links[0].entry)
         if doc_path.check():
             break
-    if not doc_path.check():
+    if doc_path is None or not doc_path.check():
         if version == 'stable':
             raise HTTPNotFound("No stable documentation available.")
         raise HTTPNotFound("No documentation available.")
     doc_path = doc_path.join(relpath)
-    if not doc_path.check():
+    if check_content and not doc_path.check():
         raise HTTPNotFound("File %s not found in documentation." % relpath)
     return dict(
         doc_path=doc_path,
@@ -99,7 +131,10 @@ def doc_serve(context, request):
     """ Serves the raw documentation files. """
     context = ContextWrapper(context)
     doc_info = get_doc_info(context, request)
-    return FileResponse(str(doc_info['doc_path']))
+    response = FileResponse(str(doc_info['doc_path']))
+    if context.version in ('latest', 'stable'):
+        response.cache_expires()
+    return response
 
 
 @view_config(
@@ -112,6 +147,24 @@ def doc_show(context, request):
     stage = context.stage
     name, version = context.project, context.version
     doc_info = get_doc_info(context, request)
+    version_links = []
+    latest_doc_info = get_doc_info(context, request, version='latest', check_content=False)
+    if latest_doc_info['doc_version'] != doc_info['doc_version']:
+        version_links.append(dict(
+            title="Latest documentation",
+            url=request.route_url(
+                "docviewroot", user=stage.user.name, index=stage.index,
+                project=name, version='latest', relpath="index.html")))
+    try:
+        stable_doc_info = get_doc_info(context, request, version='stable', check_content=False)
+        if stable_doc_info['doc_version'] != doc_info['doc_version'] and stable_doc_info['doc_version'] != latest_doc_info['doc_version']:
+            version_links.append(dict(
+                title="Stable documentation",
+                url=request.route_url(
+                    "docviewroot", user=stage.user.name, index=stage.index,
+                    project=name, version='stable', relpath="index.html")))
+    except (HTTPFound, HTTPNotFound):
+        pass
     return dict(
         title="%s-%s Documentation" % (name, version),
         base_url=request.route_url(
@@ -122,8 +175,10 @@ def doc_show(context, request):
             project=name, version=version, relpath=''),
         url=request.route_url(
             "docroot", user=stage.user.name, index=stage.index,
-            project=name, version=version, relpath=doc_info['relpath']),
+            project=name, version=version,
+            relpath=doc_info['relpath'], _query=request.query_string),
         version_mismatch=doc_info['version_mismatch'],
+        version_links=version_links,
         doc_version=doc_info['doc_version'])
 
 
@@ -135,7 +190,7 @@ def notfound(request):
     path = decode_path_info(request.environ['PATH_INFO'] or '/')
     registry = request.registry
     mapper = registry.queryUtility(IRoutesMapper)
-    if mapper is not None and path.endswith('/'):
+    if mapper is not None and path.endswith('/') and '+simple' not in path:
         # redirect URLs with a trailing slash to URLs without one, if there
         # is a matching route
         nonslashpath = path.rstrip('/')
@@ -227,9 +282,7 @@ def get_files_info(request, linkstore, show_toxresults=False):
     for link in sorted(filedata, key=attrgetter('basename')):
         url = url_for_entrypath(request, link.entrypath)
         entry = link.entry
-        if entry.eggfragment:
-            url += "#egg=%s" % entry.eggfragment
-        elif entry.hash_spec:
+        if entry.hash_spec:
             url += "#" + entry.hash_spec
         py_version, file_type = get_pyversion_filetype(link.basename)
         if py_version == 'source':
@@ -258,6 +311,7 @@ def get_files_info(request, linkstore, show_toxresults=False):
             toxresults = get_toxresults_info(linkstore, link)
             if toxresults:
                 fileinfo['toxresults'] = toxresults
+                fileinfo['toxresults_state'] = get_toxresults_state(toxresults)
         files.append(fileinfo)
     return files
 
@@ -275,6 +329,11 @@ def get_toxresults_info(linkstore, for_link, newest=True):
             log.error("Couldn't parse test results %s." % toxlink)
             continue
         for toxenv in toxenvs:
+            status = 'unknown'
+            if not toxenv.setup['failed'] and not toxenv.test['failed'] and toxenv.test['commands']:
+                status = 'passed'
+            elif toxenv.setup['failed'] or toxenv.test['failed']:
+                status = 'failed'
             info = dict(
                 basename=toxlink.basename,
                 _key="-".join(toxenv.key),
@@ -283,18 +342,32 @@ def get_toxresults_info(linkstore, for_link, newest=True):
                 envname=toxenv.envname,
                 setup=toxenv.setup,
                 test=toxenv.test,
-                failed=toxenv.failed)
+                status=status)
             if toxenv.pyversion:
                 info["pyversion"] = toxenv.pyversion
             result.append(info)
     return result
 
 
-def get_docs_info(request, stage, metadata):
+def get_toxresults_state(toxresults):
+    if not toxresults:
+        return
+    toxstates = set(x['status'] for x in toxresults)
+    if 'failed' in toxstates:
+        return 'failed'
+    if toxstates == set(['passed']):
+        return 'passed'
+    return 'unknown'
+
+
+def get_docs_info(request, stage, linkstore):
     if stage.ixconfig['type'] == 'mirror':
         return
-    name, ver = normalize_name(metadata["name"]), metadata["version"]
-    doc_path = get_unpack_path(stage, name, ver)
+    links = linkstore.get_links(rel='doczip')
+    if not links:
+        return
+    name, ver = normalize_name(linkstore.project), linkstore.version
+    doc_path = unpack_docs(stage, name, ver, links[0].entry)
     if doc_path.exists():
         return dict(
             title="%s-%s" % (name, ver),
@@ -303,6 +376,31 @@ def get_docs_info(request, stage, metadata):
                 project=name, version=ver, relpath="index.html"))
 
 
+def get_user_info(context, request, user):
+    username = user['username']
+    indexes = []
+    for index in sorted(user.get('indexes', [])):
+        stagename = "%s/%s" % (username, index)
+        stage = context.model.getstage(stagename)
+        indexes.append(dict(
+            _ixconfig=stage.ixconfig,
+            title=stagename,
+            index_name=index,
+            index_title=stage.ixconfig.get('title', None),
+            index_description=stage.ixconfig.get('description', None),
+            url=request.stage_url(stagename)))
+    return dict(
+        _user=user,
+        title=username,
+        user_name=username,
+        user_title=user.get('title', None),
+        user_description=user.get('description', None),
+        user_email=user.get('email', None),
+        user_url=request.route_url(
+            "/{user}", user=username),
+        indexes=indexes)
+
+
 @view_config(
     route_name='root',
     renderer='templates/root.pt')
@@ -312,27 +410,20 @@ def root(context, request):
         key=itemgetter('username'))
     users = []
     for user in rawusers:
-        username = user['username']
-        indexes = []
-        for index in sorted(user.get('indexes', [])):
-            stagename = "%s/%s" % (username, index)
-            stage = context.model.getstage(stagename)
-            indexes.append(dict(
-                _ixconfig=stage.ixconfig,
-                title=stagename,
-                index_name=index,
-                index_title=stage.ixconfig.get('title', None),
-                index_description=stage.ixconfig.get('description', None),
-                url=request.stage_url(stagename)))
-        users.append(dict(
-            _user=user,
-            title=username,
-            user_name=username,
-            user_title=user.get('title', None),
-            user_description=user.get('description', None),
-            user_email=user.get('email', None),
-            indexes=indexes))
-    return dict(users=users)
+        users.append(get_user_info(context, request, user))
+    return dict(
+        _context=context,
+        users=users)
+
+
+@view_config(
+    route_name="/{user}", accept="text/html", request_method="GET",
+    renderer="templates/user.pt")
+def user_get(context, request):
+    user = context.user.get()
+    return dict(
+        _context=context,
+        user=get_user_info(context, request, user))
 
 
 @view_config(
@@ -346,6 +437,7 @@ def index_get(context, request):
     packages = []
     whitelist = []
     result = dict(
+        _context=context,
         title="%s index" % stage.name,
         simple_index_url=request.simpleindex_url(stage),
         permissions=permissions,
@@ -359,7 +451,7 @@ def index_get(context, request):
         return result
 
     if hasattr(stage, "ixconfig"):
-        whitelist.extend(sorted(stage.ixconfig['mirror_whitelist']))
+        whitelist.extend(sorted(stage.ixconfig.get('mirror_whitelist', [])))
         for base in stage.ixconfig["bases"]:
             bases.append(dict(
                 title=base,
@@ -410,7 +502,7 @@ def index_get(context, request):
                 user=stage.user.name, index=stage.index,
                 project=name, version=ver),
             files=get_files_info(request, linkstore, show_toxresults),
-            docs=get_docs_info(request, stage, verdata),
+            docs=get_docs_info(request, stage, linkstore),
             _version_data=verdata))
     packages.sort(key=lambda x: x["info"]["title"])
 
@@ -425,10 +517,11 @@ def project_get(context, request):
     context = ContextWrapper(context)
     try:
         releaselinks = context.stage.get_releaselinks(context.verified_project)
+        stage_versions = context.stage.list_versions_perstage(context.verified_project)
     except context.stage.UpstreamError as e:
         log.error(e.msg)
         raise HTTPBadGateway(e.msg)
-    versions = []
+    version_info = {}
     seen = set()
     for release in releaselinks:
         user, index = release.entrypath.split("/", 2)[:2]
@@ -438,24 +531,62 @@ def project_get(context, request):
         seen_key = (user, index, name, version)
         if seen_key in seen:
             continue
-        versions.append(dict(
+        version_info[version] = dict(
             index_title="%s/%s" % (user, index),
             index_url=request.stage_url(user, index),
             title=version,
             url=request.route_url(
                 "/{user}/{index}/{project}/{version}",
                 user=user, index=index, project=name, version=version),
-            _release=release))
+            docs=None,
+            _release=release)
         seen.add(seen_key)
+    user = context.username
+    index = context.stage.index
+    index_title = "%s/%s" % (user, index)
+    name = context.verified_project
+    index_url = request.stage_url(user, index)
+    for version in stage_versions:
+        linkstore = context.stage.get_linkstore_perstage(name, version)
+        docs = get_docs_info(request, context.stage, linkstore)
+        if not docs:
+            continue
+        if version not in version_info:
+            version_info[version] = dict(
+                index_title=index_title,
+                index_url=index_url,
+                title=version,
+                url=request.route_url(
+                    "/{user}/{index}/{project}/{version}",
+                    user=user, index=index, project=name, version=version),
+                docs=docs,
+                _release=None)
+        else:
+            version_info[version]['docs'] = docs
+    versions = []
+    for version in get_sorted_versions(version_info):
+        versions.append(version_info[version])
     if hasattr(context.stage, 'get_mirror_whitelist_info'):
         whitelist_info = context.stage.get_mirror_whitelist_info(context.project)
     else:
         whitelist_info = dict(
             has_mirror_base=context.stage.has_mirror_base(context.project),
             blocked_by_mirror_whitelist=None)
+    stage = context.stage
+    latest_verdata = {}
+    latest_version = stage.get_latest_version_perstage(context.project)
+    if latest_version is not None:
+        latest_verdata = stage.get_versiondata_perstage(
+            context.project, latest_version)
     return dict(
+        _context=context,
         title="%s/: %s versions" % (context.stage.name, context.project),
         blocked_by_mirror_whitelist=whitelist_info['blocked_by_mirror_whitelist'],
+        latest_version=latest_version,
+        latest_url=request.route_url(
+            "/{user}/{index}/{project}/{version}",
+            user=user, index=index, project=name, version='latest'),
+        latest_version_data=latest_verdata,
         versions=versions)
 
 
@@ -466,10 +597,11 @@ def project_get(context, request):
 def version_get(context, request):
     """ Show version for the precise stage, ignores inheritance. """
     context = ContextWrapper(context)
-    name, version = context.verified_project, context.version
+    name = context.verified_project
     stage = context.stage
+    version = context.resolved_version
     try:
-        verdata = context.get_versiondata(perstage=True)
+        verdata = context.get_versiondata(version=version, perstage=True)
     except stage.UpstreamError as e:
         log.error(e.msg)
         raise HTTPBadGateway(e.msg)
@@ -491,7 +623,7 @@ def version_get(context, request):
     show_toxresults = (stage.ixconfig['type'] != 'mirror')
     linkstore = stage.get_linkstore_perstage(name, version)
     files = get_files_info(request, linkstore, show_toxresults)
-    docs = get_docs_info(request, stage, verdata)
+    docs = get_docs_info(request, stage, linkstore)
     home_page = verdata.get("home_page")
     nav_links = []
     if docs:
@@ -523,23 +655,45 @@ def version_get(context, request):
             nav_links.append(dict(
                 title="%s page" % base.ixconfig.get("title", "Mirror"),
                 url=mirror_web_url_fmt.format(name=name)))
+    cmp_version = Version(version)
+    if context._stable_versions:
+        stable_version = Version(context._stable_versions[0])
+        url = request.route_url(
+            "/{user}/{index}/{project}/{version}",
+            user=context.username, index=context.index,
+            project=context.project, version='stable')
+        if cmp_version.is_prerelease():
+            nav_links.append(dict(
+                title="Stable version available",
+                css_class="warning",
+                url=url))
+        elif stable_version != cmp_version:
+            nav_links.append(dict(
+                title="Newer version available",
+                css_class="severe",
+                url=url))
     return dict(
+        _context=context,
         title="%s/: %s-%s metadata and description" % (stage.name, name, version),
         content=get_description(stage, name, version),
         summary=verdata.get("summary"),
+        resolved_version=version,
         nav_links=nav_links,
         infos=infos,
+        metadata_list_fields=frozenset(
+            py.xml.escape(x)
+            for x in getattr(stage, 'metadata_list_fields', ())),
         files=files,
         blocked_by_mirror_whitelist=whitelist_info['blocked_by_mirror_whitelist'],
         show_toxresults=show_toxresults,
         make_toxresults_url=functools.partial(
             request.route_url, "toxresults",
             user=context.username, index=context.index,
-            project=context.project, version=context.version),
+            project=context.project, version=version),
         make_toxresult_url=functools.partial(
             request.route_url, "toxresult",
             user=context.username, index=context.index,
-            project=context.project, version=context.version))
+            project=context.project, version=version))
 
 
 @view_config(
@@ -553,6 +707,7 @@ def toxresults(context, request):
     toxresults = get_toxresults_info(
         linkstore, linkstore.get_links(basename=basename)[0], newest=False)
     return dict(
+        _context=context,
         title="%s/: %s-%s toxresults" % (
             context.stage.name, context.project, context.version),
         toxresults=toxresults,
@@ -577,6 +732,7 @@ def toxresult(context, request):
             linkstore.get_links(basename=basename)[0], newest=False)
         if x['basename'] == toxresult]
     return dict(
+        _context=context,
         title="%s/: %s-%s toxresult %s" % (
             context.stage.name, context.project, context.version, toxresult),
         toxresults=toxresults)
@@ -593,6 +749,7 @@ def statusview(request):
         replication_errors.append(error)
         if index >= 10:
             replication_errors.append(dict(message="More than 10 replication errors."))
+            break
     _polling_replicas = status.get('polling_replicas', {})
     polling_replicas = []
     for replica_uuid in sorted(_polling_replicas):
@@ -690,6 +847,9 @@ class SearchView:
             return
         result_info = self.search_result['info']
         batch = batch_list(result_info['pagecount'], result_info['pagenum'] - 1)
+        # if params['page'] is not in batching range,
+        # URL is broken, then let's go back to first page of batching
+        current = 0
         for index, item in enumerate(batch):
             if item is None:
                 batch_links.append(dict(
@@ -763,18 +923,9 @@ class SearchView:
                 continue
             elif text_type in ('title', 'page'):
                 docs = self.get_docs(stage, data)
-                try:
-                    entry = docs[sub_data['text_path']]
-                except KeyError:
-                    highlight = (
-                        "Couldn't access documentation files for %s "
-                        "version %s on %s. This is a bug. If you find a way "
-                        "to reproduce this, please file an issue at: "
-                        "https://bitbucket.org/hpk42/devpi/issues" % (
-                            data['name'], data['doc_version'], stage.name))
-                else:
-                    text = entry['text']
-                    highlight = search_index.highlight(text, sub_hit.get('words'))
+                entry = docs[sub_data['text_path']]
+                text = entry['text']
+                highlight = search_index.highlight(text, sub_hit.get('words'))
                 title = sub_data.get('text_title', title)
                 text_path = sub_data.get('text_path')
                 if text_path:
@@ -832,10 +983,10 @@ class SearchView:
                 item['title'] = data['name']
             item['sub_hits'] = self.process_sub_hits(
                 stage, item['sub_hits'], data)
-            more_results = result['info']['collapsed_counts'][data['path']]
+            path = data['path']
+            more_results = result['info']['collapsed_counts'][path]
             if more_results:
-                new_params = dict(self.params)
-                new_params['query'] = "%s path:%s" % (self.params['query'], data['path'])
+                new_params = make_more_url_params(self.params, path)
                 item['more_url'] = self.request.route_url(
                     'search',
                     _query=new_params)
@@ -871,84 +1022,32 @@ class SearchView:
         if method != "search":
             raise ValueError("Unknown method '%s'." % method)
         if len(data) == 2:
-            query, operator = data
+            fields, operator = data
         else:
-            query = data
+            fields = data
             operator = "and"
-        log.debug("xmlrpc_search {0}".format((query, operator)))
-        operator = operator.upper()
-        if operator not in ('AND', 'OR', 'ANDNOT', 'ANDMAYBE', 'NOT'):
-            raise ValueError("Unknown operator '%s'." % operator)
-        if set(query.keys()).difference(['name', 'summary']):
-            raise ValueError("Only 'name' and 'summary' allowed in query.")
-        parts = []
-        for key, field in (('name', 'project'), ('summary', 'summary')):
-            value = query.get(key, [])
-            if len(value) == 0:
-                continue
-            elif len(value) == 1:
-                parts.append('(type:%s "%s")' % (field, value[0].replace('"', '')))
-            else:
-                raise ValueError("Only on value allowed for query.")
-        return (" %s " % operator).join(parts)
-
-    def search_index_packages(self, query):
-        search_index = self.request.registry['search_index']
-        result = search_index.query_projects(query, page=None)
-        context = ContextWrapper(self.request.context)
-        sro = dict((x.name, i) for i, x in enumerate(context.stage.sro()))
-        # first gather basic info and only get most relevant info based on
-        # stage resolution order
-        name2stage = {}
-        name2data = {}
-        for item in result['items']:
-            data = item['data']
-            path = data['path']
-            stage = self.get_stage(path)
-            if stage is None:
-                continue
-            if stage.name not in sro:
-                continue
-            name = data['name']
-            current_stage = name2stage.get(name)
-            if current_stage is None or sro[stage.name] < sro[current_stage.name]:
-                name2stage[name] = stage
-                try:
-                    score = item['score']
-                except KeyError:
-                    score = False
-                    sub_hits = item.get('sub_hits')
-                    if sub_hits:
-                        score = sub_hits[0].get('score', False)
-                name2data[name] = dict(
-                    version=data.get('version', ''),
-                    score=score)
-        # then gather more info if available and build results
-        hits = []
-        for name, stage in name2stage.items():
-            data = name2data[name]
-            version = data['version']
-            if version and is_project_cached(stage, name):
-                metadata = stage.get_versiondata(name, version)
-                version = metadata.get('version', version)
-                summary = metadata.get('summary', '')
-            else:
-                summary = ''
-            hits.append(dict(
-                name="/%s/%s" % (stage.name, name), summary=summary,
-                version=version, _pypi_ordering=data['score']))
-        return hits
+        log.debug("xmlrpc_search {0}".format((fields, operator)))
+        return dict(fields=fields, operator=operator)
 
     @view_config(
         route_name="/{user}/{index}/", request_method="POST",
-        content_type="text/xml")
+        content_type="text/xml", is_mutating=False)
     def xmlrpc_search(self):
         try:
             query = self.query_from_xmlrpc(self.request.body)
-            log.debug("xmlrpc_search {0}".format(query))
-            hits = self.search_index_packages(query)
+            search_index = self.request.registry['search_index']
+            context = ContextWrapper(self.request.context)
+            hits = search_index.query_packages(query, list(context.stage.sro()))
             response = dumps((hits,), methodresponse=1, encoding='utf-8')
         except Exception as e:
             log.exception("Error in xmlrpc_search")
             response = dumps(Fault(1, repr(e)), encoding='utf-8')
         return Response(response)
+
+
+def make_more_url_params(params, path):
+    new_params = dict(params)
+    if 'page' in new_params:
+        del new_params['page']
+    new_params['query'] = "%s path:%s" % (params['query'], path)
+    return new_params
