@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-import json
 from devpi_common.archive import zip_dict
 from devpi_server import __version__ as devpi_server_version
 from pkg_resources import parse_version
@@ -13,7 +12,7 @@ devpi_server_version = parse_version(devpi_server_version)
 
 
 def compareable_text(text):
-    return re.sub('\s+', ' ', text.strip())
+    return re.sub(r'\s+', ' ', text.strip())
 
 
 def test_root_view(testapp):
@@ -21,6 +20,7 @@ def test_root_view(testapp):
     assert r.status_code == 200
     links = r.html.select('#content a')
     assert [(compareable_text(l.text), l.attrs['href']) for l in links] == [
+        ("root", "http://localhost/root"),
         ("root/pypi PyPI", "http://localhost/root/pypi")]
 
 
@@ -30,10 +30,37 @@ def test_root_view_with_index(mapp, testapp):
     assert r.status_code == 200
     links = r.html.select('#content a')
     assert [(compareable_text(l.text), l.attrs['href']) for l in links] == [
+        ("root", "http://localhost/root"),
         ("root/pypi PyPI", "http://localhost/root/pypi"),
+        (api.user, "http://localhost/%s" % api.user),
         (api.stagename, "http://localhost/%s" % api.stagename)]
 
 
+def test_user_view_root_pypi(testapp):
+    r = testapp.get('/root', headers=dict(accept="text/html"))
+    assert r.status_code == 200
+    links = r.html.select('#content a')
+    assert [(l.text.strip(), l.attrs['href']) for l in links] == [
+        ("root/pypi PyPI", "http://localhost/root/pypi")]
+
+
+def test_user_view(mapp, testapp):
+    api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
+    r = testapp.get("/%s" % api.user, headers=dict(accept="text/html"))
+    assert r.status_code == 200
+    links = r.html.select('#content a')
+    assert [(l.text.strip(), l.attrs['href']) for l in links] == [
+        ("%s/dev" % api.user, "http://localhost/%s/dev" % api.user)]
+
+
+def test_user_not_found(testapp):
+    r = testapp.get("/blubber", headers=dict(accept="text/html"))
+    assert r.status_code == 404
+    content, = r.html.select('#content')
+    assert "no user" in compareable_text(content.text)
+    assert "blubber" in compareable_text(content.text)
+
+
 def test_index_view_root_pypi(testapp):
     r = testapp.get('/root/pypi', headers=dict(accept="text/html"))
     assert r.status_code == 200
@@ -150,7 +177,7 @@ def test_title_description(mapp, testapp):
     (userdescription,) = content.select('.user_index_list dd.user_description')
     assert compareable_text(userdescription.text) == "userdescription"
     links = content.select('a')
-    assert [x.attrs.get('title') for x in links] == [None, 'indexdescription']
+    assert [x.attrs.get('title') for x in links] == [None, None, None, 'indexdescription']
     r = testapp.xget(200, api.index, headers=dict(accept="text/html"))
     (content,) = r.html.select('#content')
     (indextitle,) = content.select('.index_title')
@@ -171,6 +198,7 @@ def test_project_view(mapp, testapp):
     assert r.status_code == 200
     links = r.html.select('#content a')
     assert [(l.text, l.attrs['href']) for l in links] == [
+        ("2.7", "http://localhost/%s/pkg-name/latest" % api.stagename),
         (api.stagename, "http://localhost/%s" % api.stagename),
         ("2.7", "http://localhost/%s/pkg-name/2.7" % api.stagename),
         (api.stagename, "http://localhost/%s" % api.stagename),
@@ -186,6 +214,7 @@ def test_project_projectname_redirect(mapp, testapp):
     r = testapp.xget(200, api.index + '/pkg_hello', headers=dict(accept="text/html"))
     links = r.html.select('#content a')
     assert [(l.text, l.attrs['href']) for l in links] == [
+        ("1.0", "http://localhost/%s/pkg-hello/latest" % api.stagename),
         (api.stagename, "http://localhost/%s" % api.stagename),
         ("1.0", "http://localhost/%s/pkg-hello/1.0" % api.stagename)]
 
@@ -224,6 +253,7 @@ def test_project_view_root_pypi(mapp, testapp, pypistage):
     assert r.status_code == 200
     links = r.html.select('#content a')
     assert [(l.text, l.attrs['href']) for l in links] == [
+        ("2.7", "http://localhost/root/pypi/pkg1/latest"),
         ("root/pypi", "http://localhost/root/pypi"),
         ("2.7", "http://localhost/root/pypi/pkg1/2.7"),
         ("root/pypi", "http://localhost/root/pypi"),
@@ -250,6 +280,7 @@ def test_project_view_root_pypi_external_link_bad_name(mapp, testapp, pypistage)
     assert r.status_code == 200
     links = r.html.select('#content a')
     assert [(l.text, l.attrs['href']) for l in links] == [
+        ("2.7", "http://localhost/root/pypi/pkg1/latest"),
         ("root/pypi", "http://localhost/root/pypi"),
         ("2.7", "http://localhost/root/pypi/pkg1/2.7"),
         ("root/pypi", "http://localhost/root/pypi"),
@@ -272,6 +303,7 @@ def test_project_view_root_and_docs(mapp, testapp, pypistage):
     r = testapp.xget(200, api.index + '/pkg1', headers=dict(accept="text/html"))
     links = r.html.select('#content a')
     assert [(l.text, l.attrs['href']) for l in links] == [
+        ("2.6", "http://localhost/user1/dev/pkg1/latest"),
         ("root/pypi", "http://localhost/root/pypi"),
         ("2.7", "http://localhost/root/pypi/pkg1/2.7"),
         ("root/pypi", "http://localhost/root/pypi"),
@@ -279,6 +311,17 @@ def test_project_view_root_and_docs(mapp, testapp, pypistage):
         ("pkg1-2.6", "http://localhost/user1/dev/pkg1/2.6/+d/index.html")]
 
 
+def test_project_view_inherited_no_versions_on_stage(mapp, testapp):
+    mapp.login_root()
+    mapp.create_index("prod")
+    mapp.use("root/prod")
+    mapp.upload_file_pypi("pkg5-1.0.tgz", b"123", "pkg5", "1.0")
+    mapp.create_index("dev", indexconfig=dict(bases="root/prod"))
+    mapp.use("root/dev")
+    pkg_url = mapp.api.index + '/pkg5'
+    testapp.xget(200, pkg_url)
+
+
 @pytest.mark.with_notifier
 def test_version_view(mapp, testapp, monkeypatch):
     import devpi_server.model
@@ -345,7 +388,49 @@ def test_version_view(mapp, testapp, monkeypatch):
 
 
 @pytest.mark.with_notifier
-def test_version_projectname_redirect(mapp, testapp):
+def test_markdown_description_without_content_type(mapp, testapp, monkeypatch):
+    api = mapp.create_and_use()
+    mapp.upload_file_pypi(
+        "pkg1-2.6.tar.gz", b"content", "pkg1", "2.6").file_url
+    mapp.set_versiondata({
+        "name": "pkg1",
+        "version": "2.6",
+        "author": "Foo Bear",
+        "description": u'# Description'.encode('utf-8')},
+        waithooks=True)
+    r = testapp.get(api.index + '/pkg1/2.6', headers=dict(accept="text/html"))
+
+    description = r.html.select('#description')
+    assert len(description) == 1
+    assert '#' in py.builtin._totext(
+        description[0].renderContents().strip(),
+        'utf-8')
+
+
+@pytest.mark.with_notifier
+@pytest.mark.skipif(devpi_server_version < parse_version("4.7.2dev"), reason="Needs Metadata 2.1 support")
+def test_markdown_description_with_content_type(mapp, testapp, monkeypatch):
+    api = mapp.create_and_use()
+    mapp.upload_file_pypi(
+        "pkg1-2.6.tar.gz", b"content", "pkg1", "2.6").file_url
+    mapp.set_versiondata({
+        "name": "pkg1",
+        "version": "2.6",
+        "author": "Foo Bear",
+        "description": u'# Description'.encode('utf-8'),
+        "description_content_type": 'text/markdown'},
+        waithooks=True)
+    r = testapp.get(api.index + '/pkg1/2.6', headers=dict(accept="text/html"))
+
+    description = r.html.select('#description')
+    assert len(description) == 1
+    assert py.builtin._totext(
+        description[0].renderContents().strip(),
+        'utf-8') == u'<h1>Description</h1>'
+
+
+@pytest.mark.with_notifier
+def test_version_projectname(mapp, testapp):
     api = mapp.create_and_use()
     mapp.set_versiondata({
         "name": "pkg_hello", "version": "1.0", "description": "foo"})
@@ -353,8 +438,32 @@ def test_version_projectname_redirect(mapp, testapp):
         "pkg-hello-1.0.whl", b"123", "pkg-hello", "1.0",
         register=False, waithooks=True)
     r = testapp.xget(200, api.index + "/pkg-hello/1.0", headers=dict(accept="text/html"))
+    assert 'pkg-hello-1.0 metadata and description' in r.text
+
+
+@pytest.mark.with_notifier
+def test_description_updated(mapp, testapp):
+    api = mapp.create_and_use()
+    mapp.set_versiondata({
+        "name": "pkg-hello", "version": "1.0", "description": "foo"})
+    r = testapp.xget(200, api.index + "/pkg-hello/1.0", headers=dict(accept="text/html"))
     description, = r.html.select('#description')
     assert '<p>foo</p>' == py.builtin._totext(description.renderContents().strip(), 'utf-8')
+    mapp.set_versiondata({
+        "name": "pkg-hello", "version": "1.0", "description": "bar"})
+    r = testapp.xget(200, api.index + "/pkg-hello/1.0", headers=dict(accept="text/html"))
+    description, = r.html.select('#description')
+    assert '<p>bar</p>' == py.builtin._totext(description.renderContents().strip(), 'utf-8')
+
+
+@pytest.mark.with_notifier
+def test_description_empty(mapp, testapp):
+    api = mapp.create_and_use()
+    mapp.set_versiondata({
+        "name": "pkg-hello", "version": "1.0"})
+    r = testapp.xget(200, api.index + "/pkg-hello/1.0", headers=dict(accept="text/html"))
+    description, = r.html.select('#description')
+    assert '<p>No description in metadata</p>' == py.builtin._totext(description.renderContents().strip(), 'utf-8')
 
 
 def test_version_not_found(mapp, testapp):
@@ -392,12 +501,10 @@ def test_version_view_root_pypi(mapp, testapp, pypistage):
                      headers=dict(accept="text/html"))
     filesinfo = [tuple(compareable_text(t.text) for t in x.findAll('td')[:3]) for x in r.html.select('.files tbody tr')]
     assert filesinfo == [('pkg1-2.6.zip Type Source', '')]
-    links = r.html.select('#content a')
-    assert [(l.text, l.attrs['href']) for l in links] == [
-        ("Simple index", "http://localhost/root/pypi/+simple/pkg1"),
-        ("PyPI page", "https://pypi.python.org/pypi/pkg1"),
-        ("pkg1-2.6.zip", "http://localhost/root/pypi/+e/https_pypi.python.org_pkg/pkg1-2.6.zip"),
-        ("https://pypi.python.org/pypi/pkg1/2.6/", "https://pypi.python.org/pypi/pkg1/2.6/")]
+    links = {l.text: l.attrs['href'] for l in r.html.select('#content a')}
+    assert links["Simple index"] == "http://localhost/root/pypi/+simple/pkg1"
+    assert links["pkg1-2.6.zip"].startswith("http://localhost/root/pypi/+e")
+    assert "/pkg1" in links["PyPI page"]
 
 
 def test_version_view_root_pypi_external_files(mapp, testapp, pypistage):
@@ -412,11 +519,11 @@ def test_version_view_root_pypi_external_files(mapp, testapp, pypistage):
     assert silink.text == "Simple index"
     assert silink.attrs["href"] == "http://localhost/root/pypi/+simple/pkg1"
     assert pypi_link.text == "PyPI page"
-    assert pypi_link.attrs["href"] == "https://pypi.python.org/pypi/pkg1"
+    assert "/pkg1" in pypi_link.attrs["href"]
     assert link1.text == "pkg1-2.7.zip"
     assert link1.attrs["href"].endswith("pkg1-2.7.zip")
-    assert link2.text == "https://pypi.python.org/pypi/pkg1/2.7/"
-    assert link2.attrs["href"] == "https://pypi.python.org/pypi/pkg1/2.7/"
+    assert link2.text.endswith("/pkg1/2.7/")
+    assert link2.attrs["href"].endswith("/pkg1/2.7/")
 
 
 @pytest.mark.with_notifier
@@ -440,6 +547,55 @@ def test_version_view_description_errors(mapp, testapp):
     assert "Unexpected section title" in description.text
 
 
+def test_version_view_latest_stable(mapp, testapp):
+    api = mapp.create_and_use()
+    mapp.set_versiondata({
+        "name": "pkg1",
+        "version": "3.0b1"})
+    mapp.set_versiondata({
+        "name": "pkg1",
+        "version": "2.6"})
+    mapp.set_versiondata({
+        "name": "pkg1",
+        "version": "2.1b2"})
+    mapp.set_versiondata({
+        "name": "pkg1",
+        "version": "2.0"})
+    r = testapp.get(api.index + '/pkg1/2.0', headers=dict(accept="text/html"))
+    links = r.html.select('.projectnavigation a')
+    assert 'Stable version available' not in "".join(x.text for x in links)
+    link = links[-1]
+    assert link.text == 'Newer version available'
+    assert link.attrs['href'] == api.index + '/pkg1/stable'
+    r = testapp.get(api.index + '/pkg1/2.1b2', headers=dict(accept="text/html"))
+    links = r.html.select('.projectnavigation a')
+    assert 'Newer version available' not in "".join(x.text for x in links)
+    link = links[-1]
+    assert link.text == 'Stable version available'
+    assert link.attrs['href'] == api.index + '/pkg1/stable'
+    r = testapp.get(api.index + '/pkg1/2.6', headers=dict(accept="text/html"))
+    links = r.html.select('.projectnavigation a')
+    assert 'version available' not in "".join(x.text for x in links)
+    r = testapp.get(api.index + '/pkg1/3.0b1', headers=dict(accept="text/html"))
+    links = r.html.select('.projectnavigation a')
+    assert 'Newer version available' not in "".join(x.text for x in links)
+    link = links[-1]
+    assert link.text == 'Stable version available'
+    assert link.attrs['href'] == api.index + '/pkg1/stable'
+    # test stable url
+    r = testapp.get(api.index + '/pkg1/stable', headers=dict(accept="text/html"))
+    assert 'pkg1-2.6 metadata and description' in r.text
+    links = r.html.select('#navigation a')
+    assert links[4].text == '2.6'
+    assert links[4].attrs['href'] == api.index + '/pkg1/2.6'
+    # test latest url
+    r = testapp.get(api.index + '/pkg1/latest', headers=dict(accept="text/html"))
+    assert 'pkg1-3.0b1 metadata and description' in r.text
+    links = r.html.select('#navigation a')
+    assert links[4].text == '3.0b1'
+    assert links[4].attrs['href'] == api.index + '/pkg1/3.0b1'
+
+
 def test_complex_name(mapp, testapp):
     from devpi_common import __version__
     import pkg_resources
@@ -461,6 +617,7 @@ def test_complex_name(mapp, testapp):
         200, api.index + '/%s' % pkgname, headers=dict(accept="text/html"))
     links = r.html.select('#content a')
     assert [(compareable_text(l.text), l.attrs['href']) for l in links] == [
+        ('0.9', 'http://localhost/user1/dev/%s/latest' % pkgname),
         ('user1/dev', 'http://localhost/user1/dev'),
         ('0.9', 'http://localhost/user1/dev/%s/0.9' % pkgname)]
 
@@ -504,94 +661,3 @@ def test_whitelist(mapp, pypistage, testapp):
     (whitelist,) = r.html.select('.whitelist')
     text = compareable_text(whitelist.text)
     assert text == "pkg1"
-
-
-@pytest.mark.with_notifier
-def test_testdata(mapp, testapp):
-    from test_devpi_server.example import tox_result_data
-    api = mapp.create_and_use()
-    mapp.set_versiondata(
-        {"name": "pkg1", "version": "2.6", "description": "foo"})
-    mapp.upload_file_pypi(
-        "pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=200, waithooks=True)
-    path, = mapp.get_release_paths("pkg1")
-    r = testapp.post(path, json.dumps(tox_result_data))
-    assert r.status_code == 200
-    r = testapp.post(path, json.dumps({"testenvs": {"py27": {}}}))
-    r = testapp.xget(200, api.index, headers=dict(accept="text/html"))
-    passed, = r.html.select('.passed')
-    assert passed.text == 'tests'
-    assert passed.attrs['href'].endswith(
-        '/user1/dev/pkg1/2.6/+toxresults/pkg1-2.6.tgz')
-    r = testapp.xget(200, api.index + '/pkg1/2.6', headers=dict(accept="text/html"))
-    toxresult, = r.html.select('tbody .toxresults')
-    links = toxresult.select('a')
-    assert len(links) == 2
-    assert 'passed' in links[0].attrs['class']
-    assert 'foo linux2 py27' in links[0].text
-    assert 'All toxresults' in links[1].text
-    r = testapp.xget(200, links[0].attrs['href'])
-    content = "\n".join([compareable_text(x.text) for x in r.html.select('.toxresult')])
-    assert "No setup performed" in content
-    assert "everything fine" in content
-    r = testapp.xget(200, links[1].attrs['href'])
-    rows = [
-        tuple(
-            compareable_text(t.text) if len(t.text.split()) < 2 else " ".join(t.text.split())
-            for t in x.findAll('td'))
-        for x in r.html.select('tbody tr')]
-    assert rows == [
-        ("pkg1-2.6.tgz.toxresult0", "foo", "linux2", "py27",
-         "", "No setup performed Tests")]
-
-
-@pytest.mark.with_notifier
-def test_testdata_notfound(mapp, testapp):
-    # make sure we have a user, index and package
-    mapp.create_and_use()
-    mapp.set_versiondata(
-        {"name": "pkg1", "version": "2.6", "description": "foo"})
-    # now get toxresults for another version
-    r = testapp.xget(
-        404,
-        '/user1/dev/pkg1/2.7/+toxresults/pkg1-2.7.tgz',
-        headers=dict(accept="text/html"))
-    assert 'pkg1-2.7 is not registered' in r.text
-
-
-@pytest.mark.with_notifier
-def test_testdata_corrupt(mapp, testapp):
-    api = mapp.create_and_use()
-    mapp.set_versiondata(
-        {"name": "pkg1", "version": "2.6", "description": "foo"})
-    mapp.upload_file_pypi(
-        "pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=200,
-        waithooks=True)
-    path, = mapp.get_release_paths("pkg1")
-    r = testapp.post(path, json.dumps({"testenvs": {"py27": {}}}))
-    assert r.status_code == 200
-    testapp.xget(200, api.index, headers=dict(accept="text/html"))
-
-
-@pytest.mark.with_notifier
-def test_testdata_missing(mapp, testapp):
-    from test_devpi_server.example import tox_result_data
-    import os
-    api = mapp.create_and_use()
-    mapp.set_versiondata(
-        {"name": "pkg1", "version": "2.6", "description": "foo"})
-    mapp.upload_file_pypi(
-        "pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=200,
-        waithooks=True)
-    path, = mapp.get_release_paths("pkg1")
-    r = testapp.post(path, json.dumps(tox_result_data))
-    assert r.status_code == 200
-    with mapp.xom.model.keyfs.transaction(write=False):
-        stage = mapp.xom.model.getstage(api.stagename)
-        link, = stage.get_releaselinks('pkg1')
-        linkstore = stage.get_linkstore_perstage(link.project, link.version)
-        toxresult_link, = linkstore.get_links(rel="toxresult", for_entrypath=link)
-        # delete the tox result file
-        os.remove(toxresult_link.entry.file_os_path())
-    r = testapp.xget(200, api.index, headers=dict(accept="text/html"))
-    assert '.toxresult' not in r.unicode_body
